# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from networking_bagpipe.agent import bagpipe_bgp_agent as agent

from networking_bagpipe.tests.unit.agent.bagpipe_bgp_agent import base


class TestServiceBGPVPNMixin(object):

    def test_update_bgpvpn_no_plugged_ports(self):
        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l2vpn=base.BGPVPN_L2_RT10).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            self.assertEqual(0, send_attach_fn.call_count,
                             "Send attach mustn't be called")

    def test_update_bgpvpn_already_plugged_ports(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11).__dict__

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l3vpn=base.BGPVPN_L3_RT100).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=base.BGPVPN_L3_RT100),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port11,
                                              self.DUMMY_VIF11,
                                              others_rts=base.BGPVPN_L3_RT100)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'],
                                     2)

            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            send_attach_fn.assert_has_calls(expected_calls, any_order=True)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'],
                                     2,
                                     agent.BGPVPN_SERVICE,
                                     agent.BGPVPN_L3,
                                     base.BGPVPN_L3_RT100)

    def test_update_bgpvpn_same_vpn_types(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__

        dummy_port10bis = base.DummyPort(base.NETWORK1, base.PORT10,
                                         bgpvpn_port=True,
                                         evpn=base.BGPVPN_L2_RT20).__dict__

        evpn_rts = ({k: base.BGPVPN_L2_RT10[k] + base.BGPVPN_L2_RT20[k]
                    for k in agent.RT_TYPES})

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l2vpn=evpn_rts).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10,
                                              others_rts=base.BGPVPN_L2_RT10)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)

            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     agent.BGPVPN_SERVICE,
                                     agent.BGPVPN_L2,
                                     evpn_rts)

    def test_update_bgpvpn_different_vpn_types(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__

        dummy_port10bis = base.DummyPort(base.NETWORK1, base.PORT10,
                                         bgpvpn_port=True,
                                         evpn=base.BGPVPN_L2_RT10,
                                         ipvpn=base.BGPVPN_L3_RT100).__dict__

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l2vpn=base.BGPVPN_L2_RT10,
                                         l3vpn=base.BGPVPN_L3_RT100).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)

            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            send_attach_fn.assert_has_calls(expected_calls)

            for bgpvpn_type, bgpvpn_rts in [
                    (agent.BGPVPN_L2, base.BGPVPN_L2_RT10),
                    (agent.BGPVPN_L3, base.BGPVPN_L3_RT100)]:
                self._check_network_info(base.NETWORK1['id'],
                                         1,
                                         agent.BGPVPN_SERVICE,
                                         bgpvpn_type,
                                         bgpvpn_rts)

    def test_delete_bgpvpn_remaining_plugged_ports(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l2vpn=base.BGPVPN_L2_RT10).__dict__

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'],
                                     2,
                                     agent.BGPVPN_SERVICE,
                                     agent.BGPVPN_L2,
                                     base.BGPVPN_L2_RT10)

            self.agent.delete_bgpvpn(None, dummy_bgpvpn1)

            send_detach_fn.assert_has_calls(expected_calls, any_order=True)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'],
                                     2)

    def test_delete_bgpvpn_remaining_plugged_ports_after_update(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__

        update_rts = ({k: base.BGPVPN_L2_RT10[k] + base.BGPVPN_L2_RT20[k]
                       for k in agent.RT_TYPES})

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l2vpn=update_rts).__dict__

        dummy_bgpvpn1bis = base.DummyBGPVPN(base.NETWORK1,
                                            l2vpn=base.BGPVPN_L2_RT20).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=base.BGPVPN_L2_RT20),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port11,
                                              self.DUMMY_VIF11,
                                              others_rts=base.BGPVPN_L2_RT20),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'],
                                     2,
                                     agent.BGPVPN_SERVICE,
                                     agent.BGPVPN_L2,
                                     base.BGPVPN_L2_RT10)

            send_attach_fn.reset_mock()

            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            self.agent.delete_bgpvpn(None, dummy_bgpvpn1bis)

            send_attach_fn.assert_has_calls(expected_calls, any_order=True)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'],
                                     2,
                                     agent.BGPVPN_SERVICE,
                                     agent.BGPVPN_L2,
                                     base.BGPVPN_L2_RT10)

    def test_delete_bgpvpn_no_plugged_ports(self):
        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l2vpn=base.BGPVPN_L2_RT10).__dict__

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            self.agent.delete_bgpvpn(None, dummy_bgpvpn1)

            self.assertEqual(0, send_detach_fn.call_count,
                             "Send detach mustn't be called")

    def test_delete_bgpvpn_had_plugged_ports(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L2_RT10).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L2_RT10).__dict__
        dummy_detach11 = dict(id=base.PORT11['id'],
                              network_id=base.NETWORK1['id'])

        dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                         l3vpn=base.BGPVPN_L2_RT10).__dict__

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)
            self.agent.bgpvpn_port_detach(None, dummy_detach11)

            self.assertEqual(2, send_detach_fn.call_count,
                             "Send detach must be called twice")

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)

            send_detach_fn.reset_mock()
            self.agent.delete_bgpvpn(None, dummy_bgpvpn1)

            self.assertEqual(0, send_detach_fn.call_count,
                             "Send detach isn't be called")

    def _test_bgpvpn_attach_single_port(self, bgpvpn_type, bgpvpn_rts):
        bgpvpn_info = {agent.BGPVPN_TYPES_MAP[bgpvpn_type]: bgpvpn_rts}
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      **bgpvpn_info).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(bgpvpn_type,
                                              dummy_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     agent.BGPVPN_SERVICE,
                                     bgpvpn_type,
                                     bgpvpn_rts)

    def test_bgpvpn_attach_single_port_l3_bgpvpn(self):
        self._test_bgpvpn_attach_single_port(agent.BGPVPN_L3,
                                             base.BGPVPN_L3_RT100)

    def test_bgpvpn_attach_single_port_bgpvpn(self):
        self._test_bgpvpn_attach_single_port(agent.BGPVPN_L2,
                                             base.BGPVPN_L2_RT10)

    def test_bgpvpn_attach_same_port_different_bgpvpn(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10).__dict__
        dummy_port10bis = base.DummyPort(base.NETWORK1, base.PORT10,
                                         bgpvpn_port=True,
                                         evpn=base.BGPVPN_L2_RT10,
                                         ipvpn=base.BGPVPN_L3_RT100).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port10bis,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)

            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     agent.BGPVPN_SERVICE,
                                     agent.BGPVPN_L2,
                                     base.BGPVPN_L2_RT10)

            self.agent.bgpvpn_port_attach(None, dummy_port10bis)

            send_attach_fn.assert_has_calls(expected_calls)

            for bgpvpn_type, bgpvpn_rts in [
                    (agent.BGPVPN_L2, base.BGPVPN_L2_RT10),
                    (agent.BGPVPN_L3, base.BGPVPN_L3_RT100)]:
                self._check_network_info(base.NETWORK1['id'],
                                         1,
                                         agent.BGPVPN_SERVICE,
                                         bgpvpn_type,
                                         bgpvpn_rts)

    def test_bgpvpn_attach_single_port_multiple_bgpvpns(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)

            send_attach_fn.assert_has_calls(expected_calls)

            for bgpvpn_type, bgpvpn_rts in [
                    (agent.BGPVPN_L2, base.BGPVPN_L2_RT10),
                    (agent.BGPVPN_L3, base.BGPVPN_L3_RT100)]:
                self._check_network_info(base.NETWORK1['id'],
                                         1,
                                         agent.BGPVPN_SERVICE,
                                         bgpvpn_type,
                                         bgpvpn_rts)

    def test_bgpvpn_attach_multiple_ports_same_bgpvpn(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_info(base.NETWORK1['id'],
                                     2,
                                     agent.BGPVPN_SERVICE,
                                     agent.BGPVPN_L3,
                                     base.BGPVPN_L3_RT100)

    def test_bgpvpn_attach_multiple_ports_different_bgpvpns(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT200).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port20,
                                              self.DUMMY_VIF20)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port20)

            send_attach_fn.assert_has_calls(expected_calls)

            for network_id, bgpvpn_rts in [
                    (base.NETWORK1['id'], base.BGPVPN_L3_RT100),
                    (base.NETWORK2['id'], base.BGPVPN_L3_RT200)]:
                self._check_network_info(network_id,
                                         1,
                                         agent.BGPVPN_SERVICE,
                                         agent.BGPVPN_L3,
                                         bgpvpn_rts)

    def _test_bgpvpn_detach_single_port(self, bgpvpn_type, bgpvpn_rts):
        bgpvpn_info = {agent.BGPVPN_TYPES_MAP[bgpvpn_type]: bgpvpn_rts}
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      **bgpvpn_info).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(bgpvpn_type,
                                              dummy_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)

    def test_bgpvpn_detach_single_port_l3_bgpvpn(self):
        self._test_bgpvpn_detach_single_port(agent.BGPVPN_L3,
                                             base.BGPVPN_L3_RT100)

    def test_bgpvpn_detach_single_port_bgpvpn(self):
        self._test_bgpvpn_detach_single_port(agent.BGPVPN_L2,
                                             base.BGPVPN_L2_RT10)

    def test_bgpvpn_detach_single_port_multiple_bgpvpns(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      evpn=base.BGPVPN_L2_RT10,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              dummy_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)

    def test_bgpvpn_detach_multiple_ports_same_bgpvpn(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach11 = dict(id=base.PORT11['id'],
                              network_id=base.NETWORK1['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            # Attach 2 ports on BGP VPN L3 1
            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port11)

            # Detach 1 port from BGP VPN L3 1
            self.agent.bgpvpn_port_detach(None, dummy_detach10)

            # Detach remaining port from BGP VPN L3 1
            self.agent.bgpvpn_port_detach(None, dummy_detach11)

            # Check if calls on BaGPipe BGP API are as expected
            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)

    def test_bgpvpn_detach_multiple_ports_different_bgpvpns(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                      bgpvpn_port=True,
                                      ipvpn=base.BGPVPN_L3_RT200).__dict__
        dummy_detach20 = dict(id=base.PORT20['id'],
                              network_id=base.NETWORK2['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              dummy_port20,
                                              self.DUMMY_VIF20)
            ]

            self.agent.bgpvpn_port_attach(None, dummy_port10)
            self.agent.bgpvpn_port_attach(None, dummy_port20)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)
            self.agent.bgpvpn_port_detach(None, dummy_detach20)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            for network_id in [base.NETWORK1['id'], base.NETWORK2['id']]:
                self._check_network_info(network_id, 0)

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)


class TestServiceBGPVPNLinuxBridge(base.BaseTestBaGPipeBGPAgentLinuxBridge,
                                   TestServiceBGPVPNMixin):
    pass


class TestServiceBGPVPNOVS(base.BaseTestBaGPipeBGPAgentOVS,
                           TestServiceBGPVPNMixin):

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_update_bgpvpn_already_plugged_ports(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestServiceBGPVPNOVS,
                  self).test_update_bgpvpn_already_plugged_ports()
            self.assertEqual(1, gw_redir_fn.call_count)

    def test_update_bgpvpn_same_vpn_types(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestServiceBGPVPNOVS,
                  self).test_update_bgpvpn_same_vpn_types()

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_update_bgpvpn_different_vpn_types(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestServiceBGPVPNOVS,
                  self).test_update_bgpvpn_different_vpn_types()
            self.assertEqual(1, gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_delete_bgpvpn_remaining_plugged_ports(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestServiceBGPVPNOVS,
                  self).test_delete_bgpvpn_remaining_plugged_ports()

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_delete_bgpvpn_had_plugged_ports(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestServiceBGPVPNOVS,
                  self).test_delete_bgpvpn_had_plugged_ports()

            self.assertEqual(2, gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def _test_bgpvpn_attach_single_port(self, bgpvpn, network,
                                        gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestServiceBGPVPNOVS,
                  self)._test_bgpvpn_attach_single_port(bgpvpn, network)
            self.assertEqual(1 if bgpvpn == agent.BGPVPN_L3 else 0,
                             gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_bgpvpn_attach_same_port_different_bgpvpn(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestServiceBGPVPNOVS,
                  self).test_bgpvpn_attach_same_port_different_bgpvpn()
            self.assertEqual(1, gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_bgpvpn_attach_single_port_multiple_bgpvpns(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestServiceBGPVPNOVS,
                  self).test_bgpvpn_attach_single_port_multiple_bgpvpns()
            self.assertEqual(1, gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_bgpvpn_attach_multiple_ports_same_bgpvpn(self, gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestServiceBGPVPNOVS,
                  self).test_bgpvpn_attach_multiple_ports_same_bgpvpn()
            self.assertEqual(2, gw_redir_fn.call_count)

    @mock.patch('networking_bagpipe.agent.bagpipe_bgp_agent'
                '.BaGPipeBGPAgent._enable_gw_redirect',
                autospec=True)
    def test_bgpvpn_attach_multiple_ports_different_bgpvpns(self,
                                                            gw_redir_fn):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF20]):
            super(TestServiceBGPVPNOVS,
                  self).test_bgpvpn_attach_multiple_ports_different_bgpvpns()
            self.assertEqual(2, gw_redir_fn.call_count)

    def _test_bgpvpn_detach_single_port(self, bgpvpn, network):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestServiceBGPVPNOVS,
                  self)._test_bgpvpn_detach_single_port(bgpvpn, network)

    def test_bgpvpn_detach_single_port_multiple_bgpvpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               return_value=self.DUMMY_VIF10):
            super(TestServiceBGPVPNOVS,
                  self).test_bgpvpn_detach_single_port_multiple_bgpvpns()

    def test_bgpvpn_detach_multiple_ports_same_bgpvpn(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestServiceBGPVPNOVS,
                  self).test_bgpvpn_detach_multiple_ports_same_bgpvpn()

    def test_bgpvpn_detach_multiple_ports_different_bgpvpns(self):
        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF20]):
            super(TestServiceBGPVPNOVS,
                  self).test_bgpvpn_detach_multiple_ports_different_bgpvpns()

    # Test fallback and ARP gateway voodoo

    def test_fallback(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]):
            super(TestServiceBGPVPNOVS,
                  self).test_update_bgpvpn_already_plugged_ports()

            port10_with_gw_mac = base.DummyPort(base.NETWORK1,
                                                base.PORT10).__dict__
            port10_with_gw_mac.update({'gateway_mac': GW_MAC})
            port10_with_gw_mac.update(
                base.DummyBGPVPN(base.NETWORK1,
                                 l3vpn=base.BGPVPN_L3_RT100).__dict__)
            with mock.patch.object(self.agent, '_send_attach_local_port') as\
                    send_attach_fn:
                self.agent.bgpvpn_port_attach(None,
                                              copy.copy(port10_with_gw_mac))

                fallback = {'dst_mac': GW_MAC,
                            'ovs_port_number': base.PATCH_MPLS_TO_INT_OFPORT,
                            'src_mac': '00:00:5e:2a:10:00'}

                expected_calls = [
                    self._mock_send_expected_call(agent.BGPVPN_L3,
                                                  port10_with_gw_mac,
                                                  self.DUMMY_VIF10,
                                                  fallback=fallback),
                ]

                send_attach_fn.assert_has_calls(expected_calls)

    def test_gateway_arp_voodoo(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent.int_br,
                                  'add_flow') as add_flow, \
                mock.patch.object(self.agent.tun_br,
                                  'delete_flows') as tun_delete_flows,\
                mock.patch.object(self.agent.int_br,
                                  'delete_flows') as int_delete_flows:
            super(TestServiceBGPVPNOVS,
                  self).test_update_bgpvpn_already_plugged_ports()

            port10_with_gw_mac = base.DummyPort(base.NETWORK1,
                                                base.PORT10).__dict__
            port10_with_gw_mac.update({'gateway_mac': GW_MAC})
            port10_with_gw_mac.update(
                base.DummyBGPVPN(base.NETWORK1,
                                 l3vpn=base.BGPVPN_L3_RT100).__dict__)

            self.agent.bgpvpn_port_attach(None, copy.copy(port10_with_gw_mac))

            self.assertEqual(2, add_flow.call_count)

            add_flow.assert_has_calls([
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x2,
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC,
                          arp_spa='10.0.0.1',
                          actions="drop"),
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x01,
                          dl_src=GW_MAC,
                          arp_spa='10.0.0.1',
                          arp_sha=GW_MAC,
                          actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                          )
            ])

            self.agent.bgpvpn_port_detach(None,
                                          base.DummyPort(base.NETWORK1,
                                                         base.PORT10).__dict__)

            self.assertEqual(0, tun_delete_flows.call_count)
            self.assertEqual(0, int_delete_flows.call_count)

            self.agent.bgpvpn_port_detach(None,
                                          base.DummyPort(base.NETWORK1,
                                                         base.PORT11).__dict__)

            self.assertEqual(1, tun_delete_flows.call_count)
            self.assertEqual(1, int_delete_flows.call_count)

            tun_delete_flows.assert_has_calls([
                mock.call(table=mock.ANY,
                          priority=2,
                          strict=True,
                          proto='arp',
                          arp_op=0x01,
                          arp_tpa='10.0.0.1',
                          dl_vlan=mock.ANY,
                          )])
            int_delete_flows.assert_has_calls([
                mock.call(table=mock.ANY,
                          proto='arp',
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC),
            ])

    def test_gateway_arp_voodoo_update_bgpvpn_after_plug(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent.int_br,
                                  'add_flow') as add_flow:
            port10 = base.DummyPort(base.NETWORK1, base.PORT10).__dict__
            port10.update({'gateway_mac': GW_MAC})

            self.agent.bgpvpn_port_attach(None, copy.copy(port10))

            dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                             l3vpn=base.BGPVPN_L3_RT100,
                                             gateway_mac=GW_MAC).__dict__

            with mock.patch.object(self.agent, '_send_attach_local_port') as\
                    send_attach_fn:

                self.agent.update_bgpvpn(None, copy.copy(dummy_bgpvpn1))

                self.assertEqual(2, add_flow.call_count)

                add_flow.assert_has_calls([
                    mock.call(table=mock.ANY,
                              priority=2,
                              proto='arp',
                              arp_op=0x2,
                              dl_src=GW_MAC,
                              arp_sha=GW_MAC,
                              arp_spa='10.0.0.1',
                              actions="drop"),
                    mock.call(table=mock.ANY,
                              priority=2,
                              proto='arp',
                              arp_op=0x01,
                              dl_src=GW_MAC,
                              arp_spa='10.0.0.1',
                              arp_sha=GW_MAC,
                              actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                              )
                ])

                fallback = {'dst_mac': GW_MAC,
                            'ovs_port_number': base.PATCH_MPLS_TO_INT_OFPORT,
                            'src_mac': '00:00:5e:2a:10:00'}

                expected_calls = [
                    self._mock_send_expected_call(
                        agent.IPVPN,
                        port10,
                        self.DUMMY_VIF10,
                        others_rts=base.BGPVPN_L3_RT100,
                        fallback=fallback),
                ]

                send_attach_fn.assert_has_calls(expected_calls)

                self.agent.delete_bgpvpn(None, copy.copy(dummy_bgpvpn1))

                add_flow.reset_mock()
                self.assertEqual(0, add_flow.call_count)

                self.agent.update_bgpvpn(None, copy.copy(dummy_bgpvpn1))

                self.assertEqual(2, add_flow.call_count)

    def test_gateway_plug_before_update(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10]), \
                mock.patch.object(self.agent.int_br,
                                  'add_flow') as add_flow:
            port10 = base.DummyPort(base.NETWORK1, base.PORT10).__dict__

            self.agent.bgpvpn_port_attach(None, copy.copy(port10))

            dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                             l3vpn=base.BGPVPN_L3_RT100,
                                             gateway_mac=GW_MAC).__dict__

            with mock.patch.object(self.agent, '_send_attach_local_port') as\
                    send_attach_fn:

                self.agent.update_bgpvpn(None, copy.copy(dummy_bgpvpn1))

                self.assertEqual(2, add_flow.call_count)

                add_flow.assert_has_calls([
                    mock.call(table=mock.ANY,
                              priority=2,
                              proto='arp',
                              arp_op=0x2,
                              dl_src=GW_MAC,
                              arp_sha=GW_MAC,
                              arp_spa='10.0.0.1',
                              actions="drop"),
                    mock.call(table=mock.ANY,
                              priority=2,
                              proto='arp',
                              arp_op=0x01,
                              dl_src=GW_MAC,
                              arp_spa='10.0.0.1',
                              arp_sha=GW_MAC,
                              actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                              )
                ])

                fallback = {'dst_mac': GW_MAC,
                            'ovs_port_number': base.PATCH_MPLS_TO_INT_OFPORT,
                            'src_mac': '00:00:5e:2a:10:00'}

                expected_calls = [
                    self._mock_send_expected_call(
                        agent.IPVPN,
                        port10,
                        self.DUMMY_VIF10,
                        others_rts=base.BGPVPN_L3_RT100,
                        fallback=fallback),
                ]

                send_attach_fn.assert_has_calls(expected_calls)

                self.agent.delete_bgpvpn(None, copy.copy(dummy_bgpvpn1))

                add_flow.reset_mock()
                self.assertEqual(0, add_flow.call_count)

                self.agent.update_bgpvpn(None, dummy_bgpvpn1)

                self.assertEqual(2, add_flow.call_count)

    def test_evpn_no_gateway_arp_voodoo(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent.int_br,
                                  'add_flow') as add_flow, \
                mock.patch.object(self.agent.int_br,
                                  'delete_flows') as delete_flows:

            port10_with_gw_mac = base.DummyPort(base.NETWORK1,
                                                base.PORT10).__dict__
            port10_with_gw_mac.update({'gateway_mac': GW_MAC})
            port10_with_gw_mac.update(
                base.DummyBGPVPN(base.NETWORK1,
                                 l2vpn=base.BGPVPN_L2_RT10).__dict__)

            self.agent.bgpvpn_port_attach(None, copy.copy(port10_with_gw_mac))

            self.assertEqual(0, add_flow.call_count)

            self.agent.bgpvpn_port_detach(None,
                                          base.DummyPort(base.NETWORK1,
                                                         base.PORT10).__dict__)
            self.assertEqual(0, delete_flows.call_count)

            self.agent.bgpvpn_port_detach(None,
                                          base.DummyPort(base.NETWORK1,
                                                         base.PORT11).__dict__)

            self.assertEqual(0, delete_flows.call_count)

    def test_gateway_arp_voodoo_ovs_restart(self):
        GW_MAC = 'aa:bb:cc:dd:ee:ff'

        with mock.patch.object(self.agent.int_br, 'get_vif_port_by_id',
                               side_effect=[self.DUMMY_VIF10,
                                            self.DUMMY_VIF11]), \
                mock.patch.object(self.agent.int_br,
                                  'add_flow') as add_flow:
            port10 = base.DummyPort(base.NETWORK1, base.PORT10).__dict__
            port10.update({'gateway_mac': GW_MAC})

            self.agent.bgpvpn_port_attach(None, copy.copy(port10))

            dummy_bgpvpn1 = base.DummyBGPVPN(base.NETWORK1,
                                             l3vpn=base.BGPVPN_L3_RT100,
                                             gateway_mac=GW_MAC).__dict__

            self.agent.update_bgpvpn(None, dummy_bgpvpn1)

            self.assertEqual(2, add_flow.call_count)

            expected_calls = [
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x2,
                          dl_src=GW_MAC,
                          arp_sha=GW_MAC,
                          arp_spa='10.0.0.1',
                          actions="drop"),
                mock.call(table=mock.ANY,
                          priority=2,
                          proto='arp',
                          arp_op=0x01,
                          dl_src=GW_MAC,
                          arp_spa='10.0.0.1',
                          arp_sha=GW_MAC,
                          actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL"
                          )
            ]
            add_flow.assert_has_calls(expected_calls)

            add_flow.reset_mock()

            self.agent.ovs_restarted_bgpvpn()

            add_flow.assert_has_calls(expected_calls)



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import socket

import httplib2
import itertools
import json

from copy import deepcopy

from collections import defaultdict
from collections import namedtuple

import netaddr

from oslo_config import cfg

from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from oslo_concurrency import lockutils

from oslo_service import loopingcall

from networking_bagpipe._i18n import _

from networking_bagpipe.rpc import agent as bagpipe_agent_rpc
from networking_bagpipe.rpc.client import topics_BAGPIPE

from networking_bagpipe.agent.bgpvpn import rpc_agent as bgpvpn_agent_rpc
from networking_bagpipe.agent.bgpvpn.rpc_client import topics_BAGPIPE_BGPVPN

from neutron.agent.common import ovs_lib
from neutron.common import topics
from neutron.conf.agent import common as config

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib import exceptions as n_exc

from neutron.conf.plugins.ml2.drivers import ovs_conf

from neutron.plugins.ml2.drivers.linuxbridge.agent.linuxbridge_neutron_agent \
    import LinuxBridgeManager
from neutron.plugins.ml2.drivers.openvswitch.agent.common import constants\
    as a_const
from neutron.plugins.ml2.drivers.openvswitch.agent import ovs_neutron_agent
from neutron.plugins.ml2.drivers.openvswitch.agent import vlanmanager

LOG = logging.getLogger(__name__)


# Having this at line 231 is apparently not enough, so adding here as well:
# pylint: disable=not-callable


DEFAULT_GATEWAY_MAC = "00:00:5e:00:43:64"
FALLBACK_SRC_MAC = "00:00:5e:2a:10:00"

BAGPIPE_L2_SERVICE = 'bagpipe_l2'
BGPVPN_SERVICE = 'bgpvpn'
BAGPIPE_SERVICES = [BAGPIPE_L2_SERVICE, BGPVPN_SERVICE]

# bagpipe-bgp VPN types
EVPN = 'evpn'
IPVPN = 'ipvpn'
VPN_TYPES = [EVPN, IPVPN]

# BGPVPN service VPN types
BGPVPN_L2 = 'l2vpn'
BGPVPN_L3 = 'l3vpn'
BGPVPN_TYPES = [BGPVPN_L2, BGPVPN_L3]
# Map from BGPVPN service VPN types to bagpipe-bgp VPN types
BGPVPN_TYPES_MAP = {BGPVPN_L2: EVPN, BGPVPN_L3: IPVPN}

RT_IMPORT = 'import_rt'
RT_EXPORT = 'export_rt'
RT_TYPES = [RT_IMPORT, RT_EXPORT]

LINUXIF_PREFIX = "patch2tun"

bagpipe_bgp_opts = [
    cfg.IntOpt('ping_interval', default=10,
               help=_("The number of seconds the bagpipe-bgp client will "
                      "wait between polling for restart detection.")),
    cfg.StrOpt('bagpipe_bgp_ip', default='127.0.0.1',
               help=_("bagpipe-bgp REST service IP address.")),
    cfg.IntOpt('bagpipe_bgp_port', default=8082,
               help=_("bagpipe-bgp REST service IP port.")),
    cfg.StrOpt('mpls_bridge', default='br-mpls',
               help=_("OVS MPLS bridge to use")),
    cfg.StrOpt('tun_to_mpls_peer_patch_port', default='patch-to-mpls',
               help=_("OVS Peer patch port in tunnel bridge to MPLS bridge ")),
    cfg.StrOpt('mpls_to_tun_peer_patch_port', default='patch-from-tun',
               help=_("OVS Peer patch port in MPLS bridge to tunnel bridge ")),
    cfg.StrOpt('mpls_to_int_peer_patch_port', default='patch-mpls-to-int',
               help=_("OVS Peer patch port in MPLS bridge to int bridge ")),
    cfg.StrOpt('int_to_mpls_peer_patch_port', default='patch-int-from-mpls',
               help=_("OVS Peer patch port in int bridge to MPLS bridge ")),
]

cfg.CONF.register_opts(bagpipe_bgp_opts, "BAGPIPE")
ovs_conf.register_ovs_agent_opts()
config.register_agent_state_opts_helper(cfg.CONF)


class BGPAttachmentNotFound(n_exc.NotFound):
    message = "Local port %(local_port)s details could not be found"


class BaGPipeBGPException(n_exc.NeutronException):
    message = "An exception occurred when calling bagpipe-bgp \
               REST service: %(reason)s"


class HTTPClientBase(object):
    """An HTTP client base class"""

    def __init__(self, host="127.0.0.1", port=8082,
                 client_name="HTTP client base"):
        """Create a new HTTP client

        :param host: HTTP server IP address
        :param port: HTTP server port
        """
        self.host = host
        self.port = port
        self.client_name = client_name

    def do_request(self, method, action, body=None):
        LOG.debug("bagpipe-bgp client request: %(method)s %(action)s "
                  "[%(body)s]",
                  {'method': method, 'action': action, 'body': str(body)})

        if isinstance(body, dict):
            body = json.dumps(body)
        try:
            headers = {'User-Agent': self.client_name,
                       "Content-Type": "application/json",
                       "Accept": "application/json"}
            uri = "http://%s:%s/%s" % (self.host, self.port, action)

            http = httplib2.Http()
            response, content = http.request(uri, method, body, headers)
            LOG.debug("bagpipe-bgp returns [%(status)s:%(content)s]",
                      {'status': str(response.status), 'content': content})

            if response.status == 200:
                if content and len(content) > 1:
                    return json.loads(content)
            else:
                reason = (
                    "An HTTP operation has failed on bagpipe-bgp."
                )
                raise BaGPipeBGPException(reason=reason)
        except (socket.error, IOError) as e:
            reason = "Failed to connect to bagpipe-bgp: %s" % str(e)
            raise BaGPipeBGPException(reason=reason)

    def get(self, action):
        return self.do_request("GET", action)

    def post(self, action, body=None):
        return self.do_request("POST", action, body=body)

    def put(self, action, body=None):
        return self.do_request("PUT", action, body=body)

    def delete(self, action):
        return self.do_request("DELETE", action)


class DummyOVSAgent(ovs_neutron_agent.OVSNeutronAgent):
    # this class is used only to 'borrow' setup_entry_for_arp_reply
    # from OVSNeutronAgent
    arp_responder_enabled = True

    def __init__(self):
        pass


def has_attachement(bgpvpn_info, vpn_type):
    return (vpn_type in bgpvpn_info and (
            bgpvpn_info[vpn_type].get(RT_IMPORT) or
            bgpvpn_info[vpn_type].get(RT_EXPORT))
            )


class keydefaultdict(defaultdict):
    """Inherit defaultdict class to customize default_factory.

    Override __missing__ method to construct custom object as default_factory
    passing an argument.

    For example:
    class C(object):
        def __init__(self, value):
            self.value = value

    d = keydefaultdict(C)
    d[key] returns C(key)
    """
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            # pylint: disable=not-callable
            ret = self[key] = getattr(self, 'default_factory')(key)
            return ret


GatewayInfo = namedtuple('GatewayInfo', ['mac', 'ip'])
NO_GW_INFO = GatewayInfo(None, None)


class CommonInfo(object):

    def __init__(self, id):
        self.id = id
        self.service_infos = dict()

    def add_service_info(self, service_name, service_info):
        if not service_info:
            return

        if service_info != self.service_infos.get(service_name, {}):
            self.service_infos[service_name] = service_info


class PortInfo(CommonInfo):

    def __init__(self, port_id):
        super(PortInfo, self).__init__(port_id)

        self.ip_address = None
        self.mac_address = None
        self.network = None
        self.local_port = dict()

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.id == other.id)

    def __hash__(self):
        return hash(self.id)

    def set_local_port(self, linuxif):
        self.local_port = dict(linuxif=linuxif)

    def set_ip_mac_infos(self, ip_address, mac_address):
        self.ip_address = ip_address
        self.mac_address = mac_address

    def set_network(self, network):
        if not self.network:
            self.network = network
        else:
            LOG.warning('Network reference has already been set for port')


class NetworkInfo(CommonInfo):

    def __init__(self, network_id):
        super(NetworkInfo, self).__init__(network_id)

        self.gateway_info = NO_GW_INFO
        self.ports = set()

    def set_gateway_info(self, gateway_info):
        self.gateway_info = gateway_info


class BaGPipeBGPAgent(HTTPClientBase,
                      bagpipe_agent_rpc.BaGPipeAgentRpcCallBackMixin,
                      bgpvpn_agent_rpc.BGPVPNAgentRpcCallBackMixin
                      ):
    """Implements a BaGPipe-BGP REST client"""

    # bagpipe-bgp status
    BAGPIPEBGP_UP = 'UP'
    BAGPIPEBGP_DOWN = 'DOWN'

    def __init__(self, agent_type, connection, int_br=None, tun_br=None):

        """Create a new BaGPipe-BGP REST service client.

        :param agent_type: bagpipe-bgp agent type (Linux bridge or OVS)
        :param connection: RPC Connection
        :param int_br: OVS integration bridge
        :param tun_br: OVS tunnel bridge
        """
        super(BaGPipeBGPAgent,
              self).__init__(cfg.CONF.BAGPIPE.bagpipe_bgp_ip,
                             cfg.CONF.BAGPIPE.bagpipe_bgp_port, agent_type)

        self.agent_type = agent_type

        self.ping_interval = cfg.CONF.BAGPIPE.ping_interval

        # Store all ports level network and service informations
        self.ports_info = keydefaultdict(PortInfo)
        # Store all networks level network and service informations
        self.networks_info = keydefaultdict(NetworkInfo)

        self.bagpipe_bgp_status = self.BAGPIPEBGP_DOWN
        self.seq_num = 0

        # OVS-specific setup
        if self.agent_type == n_const.AGENT_TYPE_OVS:
            self.int_br = int_br
            self.tun_br = tun_br
            self._setup_mpls_br()

            registry.subscribe(self.ovs_restarted,
                               resources.AGENT,
                               events.OVS_RESTARTED)

        # RPC setup
        if self.agent_type == n_const.AGENT_TYPE_LINUXBRIDGE:
            connection.create_consumer(topics.get_topic_name(topics.AGENT,
                                                             topics_BAGPIPE,
                                                             topics.UPDATE,
                                                             cfg.CONF.host),
                                       [self], fanout=False)
        else:
            LOG.info("bagpipe-l2 RPCs disabled for OVS bridge")

        connection.create_consumer(topics.get_topic_name(topics.AGENT,
                                                         topics_BAGPIPE_BGPVPN,
                                                         topics.UPDATE),
                                   [self], fanout=True)
        connection.create_consumer(topics.get_topic_name(topics.AGENT,
                                                         topics_BAGPIPE_BGPVPN,
                                                         topics.UPDATE,
                                                         cfg.CONF.host),
                                   [self], fanout=False)

        # Starts a greenthread for bagpipe-bgp status polling
        self._start_bagpipe_bgp_status_polling(self.ping_interval)

        self.vlan_manager = vlanmanager.LocalVlanManager()

    def _check_bagpipe_bgp_status(self):
        """Trigger refresh on bagpipe-bgp restarts

        Check if bagpipe-bgp has restarted while sending ping request
        to detect sequence number change.
        If a restart is detected, re-send all registered attachments to
        bagpipe-bgp.
        """
        new_seq_num = self._request_ping()

        # Check bagpipe-bgp restart
        if new_seq_num != self.seq_num:
            if new_seq_num != -1:
                if self.seq_num != 0:
                    LOG.warning("bagpipe-bgp restart detected...")
                else:
                    LOG.info("bagpipe-bgp successfully detected")

                self.seq_num = new_seq_num
                self.bagpipe_bgp_status = self.BAGPIPEBGP_UP

                # Re-send all registered ports to bagpipe-bgp
                if self.ports_info:
                    LOG.info("Sending all registered ports to bagpipe-bgp")
                    LOG.debug("Registered ports list: %s", self.ports_info)
                    for port_id in self.ports_info:
                        self._do_port_plug(port_id)
                else:
                    LOG.info("No attachment to send to bagpipe-bgp")
            else:
                self.bagpipe_bgp_status = self.BAGPIPEBGP_DOWN

    def _start_bagpipe_bgp_status_polling(self, ping_interval=10):
        # Start bagpipe-bgp status polling at regular interval
        status_loop = loopingcall.FixedIntervalLoopingCall(
            self._check_bagpipe_bgp_status)
        status_loop.start(interval=ping_interval,
                          initial_delay=ping_interval)  # TM: why not zero ?

    def _get_network_port_infos(self, net_id, port_id):
        net_info = self.networks_info[net_id]
        port_info = self.ports_info[port_id]
        net_info.ports.add(port_info)
        port_info.set_network(net_info)

        return net_info, port_info

    def _remove_network_port_infos(self, net_id, port_id):
        port_info = self.ports_info.get(port_id)
        net_info = self.networks_info.get(net_id)

        if port_info:
            del self.ports_info[port_id]

        if net_info:
            net_info.ports.discard(port_info)
            if not net_info.ports:
                del self.networks_info[net_id]

    def _get_common_attach_info(self, port_info):
        """Format common attach informations

        {
            local_port: {
                linuxif: <LINUX_IF>
            },
            ip_address: <PORT_IP>,
            mac_address: <PORT_MAC>,
            gateway_ip: <NETWORK_GATEWAY_IP>
        }
        """
        net_info = port_info.network
        common_plug_details = {'local_port': port_info.local_port,
                               'ip_address': port_info.ip_address,
                               'mac_address': port_info.mac_address,
                               'gateway_ip': net_info.gateway_info.ip
                               }

        return common_plug_details

    def _compile_port_attach_info(self, port_id):
        port_info = self.ports_info.get(port_id)

        if not port_info:
            LOG.warning("No PortInfo found for port %s", port_id)
            return []

        net_info = port_info.network

        services = list(set(net_info.service_infos.keys()) |
                        set(port_info.service_infos.keys()))

        service_attach_info = {}
        for service in services:
            service_info = {}
            if net_info.service_infos.get(service, {}):
                service_info.update(deepcopy(net_info.service_infos[service]))
            if port_info.service_infos.get(service, {}):
                service_info.update(deepcopy(port_info.service_infos[service]))

            service_attach_info[service] = (
                getattr(self,
                        '_compile_%s_attach_info' % service)(service_info,
                                                             port_info)
            )

        attach_list = defaultdict(list)
        for vpn_type in VPN_TYPES:
            attach_info = {}

            for service in services:
                if vpn_type in service_attach_info[service]:
                    if not attach_info:
                        attach_info = self._get_common_attach_info(port_info)
                        attach_info.update({'vpn_instance_id': '%s_%s' %
                                            (net_info.id, vpn_type)})

                    service_vpn_info = service_attach_info[service][vpn_type]

                    if vpn_type not in attach_info:
                        attach_info.update(dict(vpn_type=vpn_type))

                    # Check if static routes
                    static_routes = service_vpn_info.pop('static_routes', [])
                    if static_routes:
                        static_info = deepcopy(attach_info)
                        static_info.update({'advertise_subnet': True})
                        for static_route in static_routes:
                            static_info['ip_address'] = static_route
                            static_info.update(service_vpn_info)

                            attach_list[vpn_type].append(static_info)

                    for rt_type in RT_TYPES:
                        if rt_type in service_vpn_info:
                            if rt_type not in attach_info:
                                attach_info[rt_type] = []

                            attach_info[rt_type] += (
                                service_vpn_info.pop(rt_type)
                            )

                    # Check if plugging an EVPN into an IPVPN
                    if vpn_type == IPVPN and EVPN in attach_list:
                        attach_info['local_port'] = {
                            EVPN: {
                                'id': '%s_evpn' % net_info.id
                            }
                        }
                        service_vpn_info.pop('local_port', {})
                    else:
                        attach_info['local_port'].update(
                            service_vpn_info.pop('local_port', {})
                        )

                    attach_info.update(service_vpn_info)

            if attach_info:
                attach_list[vpn_type].append(attach_info)

        return attach_list

    def _request_ping(self):
        """Send ping request to bagpipe-bgp to get sequence number"""
        try:
            response = self.get('ping')
            LOG.debug("bagpipe-bgp PING response received with "
                      "sequence number %s", response)
            return response
        except BaGPipeBGPException as e:
            LOG.warning(str(e))
            return -1

    def _send_attach_local_port(self, local_port_details):
        """Send local port attach request to BaGPipe-BGP if running"""
        if self.bagpipe_bgp_status is self.BAGPIPEBGP_UP:
            try:
                self.post('attach_localport', local_port_details)
                LOG.debug("Local port has been attached to bagpipe-bgp with "
                          "details %s", local_port_details)
            except BaGPipeBGPException as e:
                LOG.error("Can't attach local port on bagpipe-bgp: %s", str(e))
        else:
            LOG.debug("Local port not yet attached to bagpipe-bgp (not up)")

    def _send_detach_local_port(self, local_port_details):
        """Send local port detach request to BaGPipe-BGP if running"""
        if self.bagpipe_bgp_status is self.BAGPIPEBGP_UP:
            try:
                self.post('detach_localport', local_port_details)
                LOG.debug("Local port has been detached from bagpipe-bgp "
                          "with details %s", local_port_details)
            except BaGPipeBGPException as e:
                LOG.error("Can't detach local port from bagpipe-bgp: %s",
                          str(e))
                raise
        else:
            LOG.debug("Local port not yet detached from bagpipe-bgp (not up)")

    def _setup_mpls_br(self):
        '''Setup the MPLS bridge for bagpipe-bgp.

        Creates MPLS bridge, and links it to the integration and tunnel
        bridges using patch ports.

        :param mpls_br: the name of the MPLS bridge.
        '''
        mpls_br = cfg.CONF.BAGPIPE.mpls_bridge
        self.mpls_br = ovs_lib.OVSBridge(mpls_br)

        if not self.mpls_br.bridge_exists(mpls_br):
            LOG.error("Unable to enable MPLS on this agent, MPLS bridge "
                      "%(mpls_br)s doesn't exist. Agent terminated!",
                      {"mpls_br": mpls_br})
            exit(1)

        # patch ports for traffic from tun bridge to mpls bridge
        self.patch_tun_to_mpls_ofport = self.tun_br.add_patch_port(
            cfg.CONF.BAGPIPE.tun_to_mpls_peer_patch_port,
            cfg.CONF.BAGPIPE.mpls_to_tun_peer_patch_port)
        self.patch_mpls_to_tun_ofport = self.mpls_br.add_patch_port(
            cfg.CONF.BAGPIPE.mpls_to_tun_peer_patch_port,
            cfg.CONF.BAGPIPE.tun_to_mpls_peer_patch_port)

        # patch ports for traffic from mpls bridge to int bridge
        self.patch_mpls_to_int_ofport = self.mpls_br.add_patch_port(
            cfg.CONF.BAGPIPE.mpls_to_int_peer_patch_port,
            cfg.CONF.BAGPIPE.int_to_mpls_peer_patch_port)
        self.patch_int_to_mpls_ofport = self.int_br.add_patch_port(
            cfg.CONF.BAGPIPE.int_to_mpls_peer_patch_port,
            cfg.CONF.BAGPIPE.mpls_to_int_peer_patch_port)

        if (int(self.patch_tun_to_mpls_ofport) < 0 or
                int(self.patch_mpls_to_tun_ofport) < 0 or
                int(self.patch_int_to_mpls_ofport) < 0 or
                int(self.patch_mpls_to_int_ofport) < 0):
            LOG.error("Failed to create OVS patch port. Cannot have "
                      "MPLS enabled on this agent, since this version "
                      "of OVS does not support patch ports. "
                      "Agent terminated!")
            exit(1)

        patch_int_ofport = self.tun_br.get_port_ofport(
            cfg.CONF.OVS.tun_peer_patch_port)

        # In br-tun, redirect all traffic from VMs toward a BGPVPN
        # default gateway MAC address to the MPLS bridge.
        #
        # (priority >0 is needed or we hit the rule redirecting unicast to
        # the UCAST_TO_TUN table)
        self.tun_br.add_flow(
            table=a_const.PATCH_LV_TO_TUN,
            priority=1,
            in_port=patch_int_ofport,
            dl_dst=DEFAULT_GATEWAY_MAC,
            actions="output:%s" % self.patch_tun_to_mpls_ofport
        )

        # Redirect traffic from the MPLS bridge to br-int
        self.tun_br.add_flow(in_port=self.patch_tun_to_mpls_ofport,
                             actions="output:%s" % patch_int_ofport)

    def ovs_restarted(self, resources, event, trigger):
        self._setup_mpls_br()
        self.ovs_restarted_bgpvpn()
        # TODO(tmorin): need to handle restart on bagpipe-bgp side, in the
        # meantime after an OVS restart, restarting bagpipe-bgp is required

    @log_helpers.log_method_call
    def _enable_gw_redirect(self, vlan, gateway_ip):
        # Add ARP responder entry for default gateway in br-tun

        # We may compete with the ARP responder entry for the real MAC
        # if the router is on a network node and we are a compute node,
        # so we must add our rule with a higher priority. Using a different
        # priority also means that arp_responder will not remove our ARP
        # responding flows and we won't remove theirs.

        # NOTE(tmorin): consider adding priority to install_arp_responder
        # and then use it here

        # (mostly copy-pasted ovs_ofctl....install_arp_responder)
        actions = a_const.ARP_RESPONDER_ACTIONS % {
            'mac': netaddr.EUI(DEFAULT_GATEWAY_MAC, dialect=netaddr.mac_unix),
            'ip': netaddr.IPAddress(gateway_ip),
        }
        self.tun_br.add_flow(table=a_const.ARP_RESPONDER,
                             priority=2,  # see above
                             dl_vlan=vlan,
                             proto='arp',
                             arp_op=0x01,
                             arp_tpa='%s' % gateway_ip,
                             actions=actions)

    @log_helpers.log_method_call
    def _disable_gw_redirect(self, vlan, gateway_ip):
        # Remove ARP responder entry for default gateway in br-tun
        self.tun_br.delete_flows(
            strict=True,
            table=a_const.ARP_RESPONDER,
            priority=2,
            dl_vlan=vlan,
            proto='arp',
            arp_op=0x01,
            arp_tpa='%s' % gateway_ip)

    @log_helpers.log_method_call
    def _hide_real_gw_arp(self, vlan, gateway_info):
        # Kill ARP replies for the gateway IP coming on br-int from the real
        # router, if any.
        #
        # NOTE(tmorin): we assume that the router MAC exists only in this vlan.
        # Doing filtering based on the local vlan would be better, but
        # we can't do this in br-int because this bridge does tagging based
        # on ovs-vsctl port tags.
        self.int_br.add_flow(table=a_const.LOCAL_SWITCHING,
                             priority=2,
                             proto='arp',
                             arp_op=0x2,
                             dl_src=gateway_info.mac,
                             arp_sha=gateway_info.mac,
                             arp_spa=gateway_info.ip,
                             actions="drop")

        # ARP requests from the real gateway need to
        # have their IP address changed to hide the gateway
        # address or the VMs will use it to update their
        # ARP cache implicitly. Below we overwrite it with 0.0.0.0.
        self.int_br.add_flow(table=a_const.LOCAL_SWITCHING,
                             priority=2,
                             proto='arp',
                             arp_op=0x01,
                             dl_src=gateway_info.mac,
                             arp_spa=gateway_info.ip,
                             arp_sha=gateway_info.mac,
                             actions="load:0x0->NXM_OF_ARP_SPA[],NORMAL")

    @log_helpers.log_method_call
    def _unhide_real_gw_arp(self, vlan, gateway_mac):
        LOG.debug("unblocking ARP from real gateway for vlan %d (%s)",
                  vlan, gateway_mac)
        self.int_br.delete_flows(table=a_const.LOCAL_SWITCHING,
                                 proto='arp',
                                 dl_src=gateway_mac,
                                 arp_sha=gateway_mac)

    def _do_port_plug(self, port_id):
        """Send port attach request to bagpipe-bgp."""
        all_plug_details = self._compile_port_attach_info(port_id)

        # First plug E-VPNs because they could be plugged into IP-VPNs
        for vpn_type in [t for t in VPN_TYPES
                         if t in all_plug_details]:
            for plug_detail in all_plug_details[vpn_type]:
                self._send_attach_local_port(plug_detail)

    def _do_port_unplug(self, port_id):
        """Send port detach request to bagpipe-bgp."""
        all_unplug_details = self._compile_port_attach_info(port_id)

        # First unplug IP-VPNs because E-VPNs could be plugged into them
        for vpn_type in [t for t in VPN_TYPES[::-1]
                         if t in all_unplug_details]:
            for unplug_detail in all_unplug_details[vpn_type]:
                self._send_detach_local_port(unplug_detail)

    @log_helpers.log_method_call
    def _check_arp_voodoo_plug(self, net_info, gateway_info):

        if (self.agent_type != n_const.AGENT_TYPE_OVS):
            return

        # See if we need to update gateway redirection and gateway ARP
        # voodoo

        vlan = self.vlan_manager.get(net_info.id).vlan

        # NOTE(tmorin): can be improved, only needed on first plug...
        self._enable_gw_redirect(vlan, gateway_info.ip)

        # update real gateway ARP blocking...
        # remove old ARP blocking ?
        if net_info.gateway_info.mac is not None:
            self._unhide_real_gw_arp(vlan, net_info.gateway_info.mac)
        # add new ARP blocking ?
        if gateway_info.mac:
            self._hide_real_gw_arp(vlan, gateway_info)

    @log_helpers.log_method_call
    def _check_arp_voodoo_unplug(self, net_id):

        if (self.agent_type != n_const.AGENT_TYPE_OVS):
            return

        net_info = self.networks_info.get(net_id)

        if not net_info:
            return

        # if last port for this network, then cleanup gateway redirection
        # NOTE(tmorin): shouldn't we check for last *ipvpn* attachment?
        if len(net_info.ports) == 1:
            LOG.debug("last unplug, undoing voodoo ARP")
            # NOTE(tmorin): vlan lookup might break if port is already
            # unplugged from bridge ?
            vlan = self.vlan_manager.get(net_id).vlan
            self._disable_gw_redirect(vlan, net_info.gateway_info.ip)
            if net_info.gateway_info.mac is not None:
                self._unhide_real_gw_arp(vlan, net_info.gateway_info.mac)

    def _compile_bagpipe_l2_attach_info(self, service_info, port_info):
        attach_info = {
            EVPN: {
                RT_IMPORT: [service_info[EVPN][RT_IMPORT]],
                RT_EXPORT: [service_info[EVPN][RT_EXPORT]]
            }
        }

        attach_info[EVPN].update(dict(
            linuxbr=LinuxBridgeManager.get_bridge_name(port_info.network.id)))

        return attach_info

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-bgp-agent')
    def bagpipe_port_attach(self, context, port_bagpipe_info):
        port_id = port_bagpipe_info.pop('id')
        net_id = port_bagpipe_info.pop('network_id')

        net_info, port_info = self._get_network_port_infos(net_id, port_id)

        # Set IP and MAC adresses in PortInfo
        ip_address = port_bagpipe_info.pop('ip_address')
        mac_address = port_bagpipe_info.pop('mac_address')
        port_info.set_ip_mac_infos(ip_address, mac_address)

        # Set gateway IP address in NetworkInfo
        gateway_info = GatewayInfo(None,
                                   port_bagpipe_info.pop('gateway_ip'))
        net_info.set_gateway_info(gateway_info)

        port_info.set_local_port(
            LinuxBridgeManager.get_tap_device_name(port_id)
        )

        net_info.add_service_info(BAGPIPE_L2_SERVICE, port_bagpipe_info)

        self._do_port_plug(port_id)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-bgp-agent')
    def bagpipe_port_detach(self, context, port_bagpipe_info):
        port_id = port_bagpipe_info['id']
        net_id = port_bagpipe_info['network_id']

        LOG.debug("Detaching port %s from bagpipe-bgp", port_id)
        if self.ports_info.get(port_id):
            try:
                self._do_port_unplug(port_id)
            except BaGPipeBGPException as e:
                LOG.error("Can't detach port from bagpipe-bgp: %s", str(e))
            finally:
                self._remove_network_port_infos(net_id, port_id)
        else:
            LOG.warning("bagpipe-bgp agent inconsistent for BaGPipe L2 or "
                        "updated with another detach")

    def _is_last_bgpvpn_info(self, net_info, service_info):
        if BGPVPN_SERVICE not in net_info.service_infos:
            return

        orig_info = deepcopy(net_info.service_infos[BGPVPN_SERVICE])

        for vpn_type in BGPVPN_TYPES:
            if vpn_type in service_info:
                if vpn_type in orig_info:
                    for rt_type in RT_TYPES:
                        if rt_type in service_info[vpn_type]:
                            orig_info[vpn_type][rt_type] = list(
                                set(orig_info[vpn_type][rt_type]) -
                                set(service_info[vpn_type][rt_type]))

                            if not orig_info[vpn_type][rt_type]:
                                del(orig_info[vpn_type][rt_type])

                    if not orig_info[vpn_type]:
                        del(orig_info[vpn_type])

        return (not orig_info, orig_info)

    def _compile_bgpvpn_attach_info(self, service_info, port_info):
        attach_info = {}

        for bgpvpn_type, rt_type in list(
                itertools.product(BGPVPN_TYPES, RT_TYPES)):
            if rt_type in service_info.get(bgpvpn_type, {}):
                bagpipe_bgp_vpn_type = BGPVPN_TYPES_MAP[bgpvpn_type]
                if bagpipe_bgp_vpn_type not in attach_info:
                    attach_info[bagpipe_bgp_vpn_type] = defaultdict(list)

                attach_info[bagpipe_bgp_vpn_type][rt_type] += (
                    service_info[bgpvpn_type][rt_type]
                )

        if self.agent_type == n_const.AGENT_TYPE_OVS:
            # Add OVS VLAN information
            vlan = self.vlan_manager.get(port_info.network.id).vlan
            for vpn_type in (vt for vt in VPN_TYPES if vt in attach_info):
                attach_info[vpn_type].update({
                    'local_port': {
                        'ovs': {
                            'plugged': True,
                            'port_number': self.patch_mpls_to_tun_ofport,
                            'vlan': vlan
                        }
                    }
                })

            if has_attachement(attach_info, IPVPN):
                # Add fallback information if needed as well
                if port_info.network.gateway_info.mac:
                    attach_info[IPVPN].update({
                        'fallback': {
                            'dst_mac': port_info.network.gateway_info.mac,
                            'src_mac': FALLBACK_SRC_MAC,
                            'ovs_port_number': self.patch_mpls_to_int_ofport
                        }
                    })
        else:
            if has_attachement(attach_info, EVPN):
                attach_info[EVPN]['linuxbr'] = (
                    LinuxBridgeManager.get_bridge_name(port_info.network.id)
                )
            if has_attachement(attach_info, IPVPN):
                # the interface we need to pass to bagpipe is the
                # bridge
                attach_info[IPVPN]['local_port'] = {
                    'linuxif':
                        LinuxBridgeManager.get_bridge_name(
                            port_info.network.id)
                }
                # NOTE(tmorin): fallback support still missing

        return attach_info

    def ovs_restarted_bgpvpn(self):
        for net_info in self.networks_info.values():
            if net_info.ports and net_info.gateway_info != NO_GW_INFO:
                bgpvpn_info = net_info.service_infos.get(BGPVPN_SERVICE)
                if has_attachement(bgpvpn_info, BGPVPN_L3):
                    self._check_arp_voodoo_plug(net_info,
                                                net_info.gateway_info)

    @log_helpers.log_method_call
    def create_bgpvpn(self, context, bgpvpn):
        self.update_bgpvpn(context, bgpvpn)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-bgp-agent')
    def update_bgpvpn(self, context, bgpvpn):
        net_id = bgpvpn.pop('network_id')

        net_info = self.networks_info.get(net_id)

        if not net_info:
            return

        new_gw_info = GatewayInfo(
            bgpvpn.pop('gateway_mac', None),
            net_info.gateway_info.ip
        )

        if has_attachement(bgpvpn, BGPVPN_L3):
            self._check_arp_voodoo_plug(net_info, new_gw_info)

        net_info.set_gateway_info(new_gw_info)

        net_info.add_service_info(BGPVPN_SERVICE, bgpvpn)

        for port_info in net_info.ports:
            self._do_port_plug(port_info.id)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-bgp-agent')
    def delete_bgpvpn(self, context, bgpvpn):
        net_id = bgpvpn.pop('network_id')

        net_info = self.networks_info.get(net_id)

        if not net_info:
            return

        # Check if remaining BGPVPN informations, otherwise unplug
        # port from bagpipe-bgp
        last_bgpvpn, updated_info = (
            self._is_last_bgpvpn_info(net_info, bgpvpn)
        )

        if self.agent_type == n_const.AGENT_TYPE_OVS:
            if (last_bgpvpn or
                    not has_attachement(updated_info, BGPVPN_L3)):
                self._check_arp_voodoo_unplug(net_id)

        if last_bgpvpn:
            if len(net_info.service_infos) == 1:
                for port_info in net_info.ports:
                    self._do_port_unplug(port_info.id)

                del(net_info.service_infos[BGPVPN_SERVICE])
            else:
                del(net_info.service_infos[BGPVPN_SERVICE])

                for port_info in net_info.ports:
                    self._do_port_plug(port_info.id)
        else:
            net_info.service_infos[BGPVPN_SERVICE] = updated_info

            for port_info in net_info.ports:
                self._do_port_plug(port_info.id)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-bgp-agent')
    def bgpvpn_port_attach(self, context, port_bgpvpn_info):
        port_id = port_bgpvpn_info.pop('id')
        net_id = port_bgpvpn_info.pop('network_id')

        net_info, port_info = self._get_network_port_infos(net_id, port_id)

        # Set IP and MAC adresses in PortInfo
        ip_address = port_bgpvpn_info.pop('ip_address')
        mac_address = port_bgpvpn_info.pop('mac_address')
        port_info.set_ip_mac_infos(ip_address, mac_address)

        # Set gateway IP and MAC (if defined) addresses in NetworkInfo
        gateway_info = GatewayInfo(port_bgpvpn_info.pop('gateway_mac', None),
                                   port_bgpvpn_info.pop('gateway_ip'))

        if has_attachement(port_bgpvpn_info, BGPVPN_L3):
            self._check_arp_voodoo_plug(net_info, gateway_info)

        net_info.set_gateway_info(gateway_info)

        if self.agent_type == n_const.AGENT_TYPE_OVS:
            vlan = self.vlan_manager.get(net_id).vlan
            port_info.set_local_port('%s:%s' % (LINUXIF_PREFIX, vlan))
        else:
            port_info.set_local_port(
                LinuxBridgeManager.get_tap_device_name(port_id)
            )

        net_info.add_service_info(BGPVPN_SERVICE, port_bgpvpn_info)

        self._do_port_plug(port_id)

    @log_helpers.log_method_call
    @lockutils.synchronized('bagpipe-bgp-agent')
    def bgpvpn_port_detach(self, context, port_bgpvpn_info):
        port_id = port_bgpvpn_info['id']
        net_id = port_bgpvpn_info['network_id']

        if self.ports_info.get(port_id):
            try:
                self._do_port_unplug(port_id)
                if self.agent_type == n_const.AGENT_TYPE_OVS:
                    net_info = self.networks_info.get(net_id)
                    bgpvpn_info = net_info.service_infos.get(BGPVPN_SERVICE)

                    if has_attachement(bgpvpn_info, BGPVPN_L3):
                        self._check_arp_voodoo_unplug(net_id)
            except BaGPipeBGPException as e:
                LOG.error("Can't detach BGPVPN port from bagpipe-bgp %s",
                          str(e))
            finally:
                self._remove_network_port_infos(net_id, port_id)
        else:
            LOG.warning('bagpipe-bgp agent inconsistent for BGPVPN or updated '
                        'with another detach')



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_bagpipe.agent import bagpipe_bgp_agent as agent

from networking_bagpipe.tests.unit.agent.bagpipe_bgp_agent import base


class TestServiceCombinedMixin(object):

    def test_combined_attach_single_port_evpns(self):
        bagpipe_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                       bgpvpn_port=True,
                                       evpn=base.BGPVPN_L2_RT10).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=base.BAGPIPE_L2_RT1)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     agent.BAGPIPE_L2_SERVICE,
                                     agent.EVPN,
                                     base.BAGPIPE_L2_RT1)
            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     agent.BGPVPN_SERVICE,
                                     agent.BGPVPN_L2,
                                     base.BGPVPN_L2_RT10)

    def test_combined_attach_single_port_different_vpns(self):
        bagpipe_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                       bgpvpn_port=True,
                                       ipvpn=base.BGPVPN_L3_RT100).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     agent.BAGPIPE_L2_SERVICE,
                                     agent.EVPN,
                                     base.BAGPIPE_L2_RT1)
            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     agent.BGPVPN_SERVICE,
                                     agent.BGPVPN_L3,
                                     base.BGPVPN_L3_RT100)

    def test_combined_attach_multiple_ports_same_evpns(self):
        bagpipe_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                       bgpvpn_port=True,
                                       evpn=base.BGPVPN_L2_RT10).__dict__

        bagpipe_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                       bgpvpn_port=True,
                                       evpn=base.BGPVPN_L2_RT10).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=base.BAGPIPE_L2_RT1),
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port11,
                                              self.DUMMY_VIF11,
                                              others_rts=base.BGPVPN_L2_RT10),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              bgpvpn_port11,
                                              self.DUMMY_VIF11,
                                              others_rts=base.BAGPIPE_L2_RT1)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_attach(None, bagpipe_port11)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port11)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_info(base.NETWORK1['id'],
                                     2,
                                     agent.BAGPIPE_L2_SERVICE,
                                     agent.EVPN,
                                     base.BAGPIPE_L2_RT1)
            self._check_network_info(base.NETWORK1['id'],
                                     2,
                                     agent.BGPVPN_SERVICE,
                                     agent.BGPVPN_L2,
                                     base.BGPVPN_L2_RT10)

    def test_combined_attach_multiple_ports_different_evpns(self):
        bagpipe_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                       bgpvpn_port=True,
                                       evpn=base.BGPVPN_L2_RT10).__dict__

        bagpipe_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                        evpn=base.BAGPIPE_L2_RT2).__dict__
        bgpvpn_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                       bgpvpn_port=True,
                                       evpn=base.BGPVPN_L2_RT20).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=base.BAGPIPE_L2_RT1),
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port20,
                                              self.DUMMY_VIF20),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              bgpvpn_port20,
                                              self.DUMMY_VIF20,
                                              others_rts=base.BAGPIPE_L2_RT2)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_attach(None, bagpipe_port20)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port20)

            send_attach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            for network_id, bagpipe_rts, bgpvpn_rts in [
                    (base.NETWORK1['id'], base.BAGPIPE_L2_RT1,
                     base.BGPVPN_L2_RT10),
                    (base.NETWORK2['id'], base.BAGPIPE_L2_RT2,
                     base.BGPVPN_L2_RT20)]:
                self._check_network_info(network_id,
                                         1,
                                         agent.BAGPIPE_L2_SERVICE,
                                         agent.EVPN,
                                         bagpipe_rts)
                self._check_network_info(network_id,
                                         1,
                                         agent.BGPVPN_SERVICE,
                                         agent.BGPVPN_L2,
                                         bgpvpn_rts)

    def test_combined_attach_multiple_ports_different_vpns(self):
        bagpipe_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                       bgpvpn_port=True,
                                       ipvpn=base.BGPVPN_L3_RT100).__dict__

        bagpipe_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                        evpn=base.BAGPIPE_L2_RT2).__dict__
        bgpvpn_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                       bgpvpn_port=True,
                                       ipvpn=base.BGPVPN_L3_RT200).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True),
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port20,
                                              self.DUMMY_VIF20),
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port20,
                                              self.DUMMY_VIF20),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              bgpvpn_port20,
                                              self.DUMMY_VIF20,
                                              evpn2ipvpn=True)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_attach(None, bagpipe_port20)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port20)

            send_attach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            for network_id, bagpipe_rts, bgpvpn_rts in [
                    (base.NETWORK1['id'], base.BAGPIPE_L2_RT1,
                     base.BGPVPN_L3_RT100),
                    (base.NETWORK2['id'], base.BAGPIPE_L2_RT2,
                     base.BGPVPN_L3_RT200)]:
                self._check_network_info(network_id,
                                         1,
                                         agent.BAGPIPE_L2_SERVICE,
                                         agent.EVPN,
                                         bagpipe_rts)
                self._check_network_info(network_id,
                                         1,
                                         agent.BGPVPN_SERVICE,
                                         agent.BGPVPN_L3,
                                         bgpvpn_rts)

    def test_combined_detach_single_port_evpns1(self):
        bagpipe_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                       bgpvpn_port=True,
                                       evpn=base.BGPVPN_L2_RT10).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=base.BAGPIPE_L2_RT1)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'], 0)

            self.agent.bagpipe_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)

    def test_combined_detach_single_port_evpns2(self):
        bagpipe_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                       bgpvpn_port=True,
                                       evpn=base.BGPVPN_L2_RT10).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        with mock.patch.object(
            self.agent,
            '_send_detach_local_port',
            side_effect=[None,
                         agent.BaGPipeBGPException(reason="Port not plugged")]
        ) as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=base.BGPVPN_L2_RT10)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_detach(None, dummy_detach10)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'], 0)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)

    def test_combined_detach_single_port_different_vpns(self):
        bagpipe_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                       bgpvpn_port=True,
                                       ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True),
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)
            self.agent.bagpipe_port_attach(None, bagpipe_port10)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)
            self.agent.bagpipe_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)

    def test_combined_detach_multiple_ports_same_evpns(self):
        bagpipe_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                       bgpvpn_port=True,
                                       evpn=base.BGPVPN_L2_RT10).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        bagpipe_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                       bgpvpn_port=True,
                                       evpn=base.BGPVPN_L2_RT10).__dict__
        dummy_detach11 = dict(id=base.PORT11['id'],
                              network_id=base.NETWORK1['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              others_rts=base.BAGPIPE_L2_RT1),
                self._mock_send_expected_call(agent.BGPVPN_L2,
                                              bgpvpn_port11,
                                              self.DUMMY_VIF11,
                                              others_rts=base.BAGPIPE_L2_RT1)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_attach(None, bagpipe_port11)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port11)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)
            self.agent.bagpipe_port_detach(None, dummy_detach10)

            self.agent.bgpvpn_port_detach(None, dummy_detach11)
            self.agent.bagpipe_port_detach(None, dummy_detach11)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'], 0)

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)

    def test_combined_detach_multiple_ports_different_vpns(self):
        bagpipe_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                        evpn=base.BAGPIPE_L2_RT1).__dict__
        bgpvpn_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                       bgpvpn_port=True,
                                       ipvpn=base.BGPVPN_L3_RT100).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        bagpipe_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                        evpn=base.BAGPIPE_L2_RT2).__dict__
        bgpvpn_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                       bgpvpn_port=True,
                                       ipvpn=base.BGPVPN_L3_RT200).__dict__
        dummy_detach20 = dict(id=base.PORT20['id'],
                              network_id=base.NETWORK2['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              bgpvpn_port10,
                                              self.DUMMY_VIF10,
                                              evpn2ipvpn=True),
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.BGPVPN_L3,
                                              bgpvpn_port20,
                                              self.DUMMY_VIF20,
                                              evpn2ipvpn=True),
                self._mock_send_expected_call(agent.EVPN,
                                              bagpipe_port20,
                                              self.DUMMY_VIF20)
            ]

            self.agent.bagpipe_port_attach(None, bagpipe_port10)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port10)

            self.agent.bagpipe_port_attach(None, bagpipe_port20)
            self.agent.bgpvpn_port_attach(None, bgpvpn_port20)

            self.agent.bgpvpn_port_detach(None, dummy_detach10)
            self.agent.bagpipe_port_detach(None, dummy_detach10)

            self.agent.bgpvpn_port_detach(None, dummy_detach20)
            self.agent.bagpipe_port_detach(None, dummy_detach20)

            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            for network_id in [base.NETWORK1['id'], base.NETWORK2['id']]:
                self._check_network_info(network_id, 0)

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)


class TestServiceCombinedLinuxBridge(
        base.BaseTestBaGPipeBGPAgentLinuxBridge, TestServiceCombinedMixin):
    pass



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

from oslo_config import cfg

from oslo_utils import uuidutils

from networking_bagpipe.agent import bagpipe_bgp_agent as agent

from neutron.plugins.ml2.drivers.linuxbridge.agent.linuxbridge_neutron_agent \
    import LinuxBridgeManager
from neutron.plugins.ml2.drivers.openvswitch.agent import vlanmanager

from neutron.tests import base
from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent import (
    ovs_test_base)

from neutron_lib import constants as n_const

PATCH_TUN_TO_MPLS_OFPORT = 1
PATCH_TUN_OFPORTS = [PATCH_TUN_TO_MPLS_OFPORT]

PATCH_INT_TO_MPLS_OFPORT = 5
PATCH_INT_OFPORTS = [PATCH_INT_TO_MPLS_OFPORT]

PATCH_MPLS_TO_TUN_OFPORT = 2
PATCH_MPLS_TO_INT_OFPORT = 6
PATCH_MPLS_OFPORTS = [PATCH_MPLS_TO_TUN_OFPORT, PATCH_MPLS_TO_INT_OFPORT]

PORT10 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:be:ef',
          'ip_address': '10.0.0.2'}

PORT11 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:f0:0d',
          'ip_address': '10.0.0.3'}

NETWORK1 = {'id': uuidutils.generate_uuid(),
            'gateway_ip': '10.0.0.1'}

PORT20 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:be:ef',
          'ip_address': '20.0.0.2'}

PORT21 = {'id': uuidutils.generate_uuid(),
          'mac_address': '00:00:de:ad:f0:0d',
          'ip_address': '20.0.0.3'}

NETWORK2 = {'id': uuidutils.generate_uuid(),
            'gateway_ip': '20.0.0.1'}

LOCAL_VLAN_MAP = {
    NETWORK1['id']: 31,
    NETWORK2['id']: 52
}

BAGPIPE_L2_RT1 = {'import_rt': 'BAGPIPE_L2:1',
                  'export_rt': 'BAGPIPE_L2:1'}

BAGPIPE_L2_RT2 = {'import_rt': 'BAGPIPE_L2:2',
                  'export_rt': 'BAGPIPE_L2:2'}

BGPVPN_L2_RT10 = {'import_rt': ['BGPVPN_L2:10'],
                  'export_rt': ['BGPVPN_L2:10']}

BGPVPN_L2_RT20 = {'import_rt': ['BGPVPN_L2:20'],
                  'export_rt': ['BGPVPN_L2:20']}

BGPVPN_L3_RT100 = {'import_rt': ['BGPVPN_L3:100'],
                   'export_rt': ['BGPVPN_L3:100']}

BGPVPN_L3_RT200 = {'import_rt': ['BGPVPN_L3:200'],
                   'export_rt': ['BGPVPN_L3:200']}


class DummyPort(object):
    def __init__(self, network, port, bgpvpn_port=False,
                 evpn=None, ipvpn=None):
        self.id = port['id']
        self.network_id = network['id']
        self.mac_address = port['mac_address']
        self.ip_address = port['ip_address']
        self.gateway_ip = network['gateway_ip']

        if bgpvpn_port:
            if evpn:
                self.l2vpn = copy.deepcopy(evpn)

            if ipvpn:
                self.l3vpn = copy.deepcopy(ipvpn)
        else:
            if evpn:
                self.evpn = copy.deepcopy(evpn)

            if ipvpn:
                self.ipvpn = copy.deepcopy(ipvpn)


class DummyVif(object):
    def __init__(self, ofport, port_name):
        self.ofport = ofport
        self.port_name = port_name


class DummyBGPVPN(object):
    def __init__(self, network, l2vpn=None, l3vpn=None, gateway_mac=None):
        self.network_id = network['id']

        if l2vpn:
            self.l2vpn = copy.deepcopy(l2vpn)

        if l3vpn:
            self.l3vpn = copy.deepcopy(l3vpn)

        if gateway_mac:
            self.gateway_mac = gateway_mac


class RTList(list):

    def __eq__(self, other):
        return set(self) == set(other)


class BaseTestBaGPipeBGPAgent(object):

    DUMMY_VIF10 = None
    DUMMY_VIF11 = None
    DUMMY_VIF20 = None
    DUMMY_VIF21 = None

    def _get_expected_local_port(self, vpn_type,
                                 network_id, port_id, vif_name):
        raise NotImplementedError

    def _format_as_list(self, value):
        return value if isinstance(value, list) else [value]

    def _get_expected_route_target(self, vpn_type, port, others_rts):
        import_rt = RTList()
        export_rt = RTList()

        if vpn_type in port:
            import_rt += self._format_as_list(port[vpn_type]['import_rt'])
            export_rt += self._format_as_list(port[vpn_type]['export_rt'])

        if others_rts:
            import_rt += self._format_as_list(others_rts.get('import_rt'))
            export_rt += self._format_as_list(others_rts.get('export_rt'))

        return import_rt, export_rt

    def _mock_send_expected_call(self, vpn_type, port, vif, evpn2ipvpn=False,
                                 others_rts=None, fallback=None):
        network_id = port['network_id']

        vif_name = vif.port_name if vif else None
        local_port, linuxbr = self._get_expected_local_port(vpn_type,
                                                            network_id,
                                                            port['id'],
                                                            vif_name)
        # Change local port if plugging evpn into ipvpn
        if evpn2ipvpn:
            local_port = dict(evpn=dict(id=network_id + '_evpn'))

        import_rt, export_rt = self._get_expected_route_target(vpn_type,
                                                               port,
                                                               others_rts)

        if vpn_type in agent.BGPVPN_TYPES:
            vpn_type = agent.BGPVPN_TYPES_MAP[vpn_type]

        expected_call = dict(vpn_instance_id=network_id + '_' + vpn_type,
                             vpn_type=vpn_type,
                             local_port=local_port,
                             mac_address=port['mac_address'],
                             ip_address=port['ip_address'],
                             gateway_ip=port['gateway_ip'],
                             import_rt=import_rt,
                             export_rt=export_rt)

        if linuxbr:
            expected_call.update(dict(
                linuxbr=LinuxBridgeManager.get_bridge_name(network_id))
            )
        if fallback:
            expected_call.update({'fallback': fallback})

        return mock.call(expected_call)

    def _check_network_info(self, network_id, expected_size,
                            service=None, vpn_type=None, vpn_rts=None):
        if expected_size == 0:
            self.assertNotIn(network_id, self.agent.networks_info,
                             "Network %s expected to have no ports left"
                             % network_id)
        else:
            self.assertIn(network_id, self.agent.networks_info)
            network_info = self.agent.networks_info[network_id]
            self.assertEqual(len(network_info.ports), expected_size,
                             "Network ports size not as expected")

            if service and vpn_type and vpn_rts:
                service_infos = network_info.service_infos
                self.assertIn(service, service_infos,
                              "No %s service details found for network %s" %
                              (service, network_id))

                self.assertIn(vpn_type, service_infos[service],
                              "No %s service %s details found for "
                              "network %s" % (service,
                                              vpn_type,
                                              network_id))

                self.assertEqual(vpn_rts, service_infos[service][vpn_type],
                                 "%s service %s details not matching %s for "
                                 "network %s" % (service,
                                                 vpn_type,
                                                 vpn_rts,
                                                 network_id))


class BaseTestBaGPipeBGPAgentLinuxBridge(base.BaseTestCase,
                                         BaseTestBaGPipeBGPAgent):

    def setUp(self):
        cfg.CONF.set_override('enable_vxlan', False, 'VXLAN')
        super(BaseTestBaGPipeBGPAgentLinuxBridge, self).setUp()

        self.bridge_mappings = {}
        self.interface_mappings = {}

        self.agent = agent.BaGPipeBGPAgent(n_const.AGENT_TYPE_LINUXBRIDGE,
                                           mock.Mock())

    def _get_expected_local_port(self, vpn_type,
                                 network_id, port_id, vif_name):
        if vpn_type == agent.EVPN or vpn_type == agent.BGPVPN_L2:
            local_port = dict(
                linuxif=LinuxBridgeManager.get_tap_device_name(port_id)
            )
            linuxbr = LinuxBridgeManager.get_bridge_name(network_id)
        elif vpn_type == agent.BGPVPN_L3:
            local_port = dict(
                linuxif=LinuxBridgeManager.get_bridge_name(network_id)
            )
            linuxbr = None
        else:
            raise Exception("not supported by current test suite: %s",
                            vpn_type)

        return local_port, linuxbr


class BaseTestBaGPipeBGPAgentOVS(ovs_test_base.OVSOFCtlTestBase,
                                 BaseTestBaGPipeBGPAgent):

    DUMMY_VIF10 = DummyVif(10, 'VIF10')
    DUMMY_VIF11 = DummyVif(11, 'VIF11')
    DUMMY_VIF20 = DummyVif(20, 'VIF20')
    DUMMY_VIF21 = DummyVif(21, 'VIF21')

    def setUp(self):
        super(BaseTestBaGPipeBGPAgentOVS, self).setUp()

        self.mock_int_br = mock.Mock()
        self.mock_int_br.add_patch_port = mock.Mock()
        self.mock_int_br.add_patch_port.side_effect = PATCH_INT_OFPORTS
        self.mock_int_br.get_vif_port_by_id = mock.Mock()

        self.mock_tun_br = mock.Mock()
        self.mock_tun_br.add_patch_port = mock.Mock()
        self.mock_tun_br.add_patch_port.side_effect = PATCH_TUN_OFPORTS
        self.mock_tun_br.get_port_ofport = mock.Mock()

        with mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                        'bridge_exists', return_value=True), \
                mock.patch('neutron.agent.common.ovs_lib.OVSBridge.'
                           'add_patch_port', side_effect=PATCH_MPLS_OFPORTS):
            self.agent = agent.BaGPipeBGPAgent(n_const.AGENT_TYPE_OVS,
                                               mock.Mock(),
                                               int_br=self.mock_int_br,
                                               tun_br=self.mock_tun_br)

        self.vlan_manager = vlanmanager.LocalVlanManager()
        for net_id, vlan in LOCAL_VLAN_MAP.items():
            try:
                self.vlan_manager.add(net_id, vlan, None, None, None)
            except vlanmanager.MappingAlreadyExists:
                pass

    def _get_expected_local_port(self, vpn_type,
                                 network_id, port_id, vif_name):
        vlan = self.vlan_manager.get(network_id).vlan
        local_port = dict(
            linuxif="patch2tun:%s" % vlan,
            ovs=dict(plugged=True,
                     port_number=PATCH_MPLS_TO_TUN_OFPORT,
                     vlan=vlan)
        )

        return local_port, None



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_bagpipe.agent import bagpipe_bgp_agent as agent

from networking_bagpipe.tests.unit.agent.bagpipe_bgp_agent import base


class TestServiceBaGPipeL2Mixin(object):

    def test_bagpipe_l2_attach_single_port(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bagpipe_port_attach(None, dummy_port10)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     agent.BAGPIPE_L2_SERVICE,
                                     agent.EVPN,
                                     base.BAGPIPE_L2_RT1)

    def test_bagpipe_l2_attach_multiple_ports_same_network(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_attach(None, dummy_port11)

            send_attach_fn.assert_has_calls(expected_calls)

            self._check_network_info(base.NETWORK1['id'],
                                     2,
                                     agent.BAGPIPE_L2_SERVICE,
                                     agent.EVPN,
                                     base.BAGPIPE_L2_RT1)

    def test_bagpipe_l2_attach_multiple_ports_different_networks(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                      evpn=base.BAGPIPE_L2_RT2).__dict__

        with mock.patch.object(self.agent,
                               '_send_attach_local_port') as send_attach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port20,
                                              self.DUMMY_VIF20)
            ]

            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_attach(None, dummy_port20)

            send_attach_fn.assert_has_calls(expected_calls)

            for network_id, evpn_rt in [
                    (base.NETWORK1['id'], base.BAGPIPE_L2_RT1),
                    (base.NETWORK2['id'], base.BAGPIPE_L2_RT2)]:
                self._check_network_info(network_id,
                                         1,
                                         agent.BAGPIPE_L2_SERVICE,
                                         agent.EVPN,
                                         evpn_rt)

    def test_bagpipe_l2_detach_single_port(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10)
            ]

            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_detach(None, dummy_detach10)

            send_detach_fn.assert_has_calls(expected_calls)

            self._check_network_info(base.NETWORK1['id'], 0)
            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)

    def test_bagpipe_l2_detach_multiple_ports_same_network(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_detach11 = dict(id=base.PORT11['id'],
                              network_id=base.NETWORK1['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port11,
                                              self.DUMMY_VIF11)
            ]

            # Attach 2 ports on network 1
            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_attach(None, dummy_port11)

            # Detach 1 port from network 1
            self.agent.bagpipe_port_detach(None, dummy_detach10)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'],
                                     1,
                                     agent.BAGPIPE_L2_SERVICE,
                                     agent.EVPN,
                                     base.BAGPIPE_L2_RT1)

            # Detach remaining port from network 1
            self.agent.bagpipe_port_detach(None, dummy_detach11)

            # Check if calls on BaGPipe BGP API are as expected
            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            self._check_network_info(base.NETWORK1['id'], 0)
            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)

    def test_bagpipe_l2_detach_multiple_ports_different_networks(self):
        dummy_port10 = base.DummyPort(base.NETWORK1, base.PORT10,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_detach10 = dict(id=base.PORT10['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port11 = base.DummyPort(base.NETWORK1, base.PORT11,
                                      evpn=base.BAGPIPE_L2_RT1).__dict__
        dummy_detach11 = dict(id=base.PORT11['id'],
                              network_id=base.NETWORK1['id'])

        dummy_port20 = base.DummyPort(base.NETWORK2, base.PORT20,
                                      evpn=base.BAGPIPE_L2_RT2).__dict__
        dummy_detach20 = dict(id=base.PORT20['id'],
                              network_id=base.NETWORK2['id'])

        dummy_port21 = base.DummyPort(base.NETWORK2, base.PORT21,
                                      evpn=base.BAGPIPE_L2_RT2).__dict__
        dummy_detach21 = dict(id=base.PORT21['id'],
                              network_id=base.NETWORK2['id'])

        with mock.patch.object(self.agent,
                               '_send_detach_local_port') as send_detach_fn:
            expected_calls = [
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port10,
                                              self.DUMMY_VIF10),
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port20,
                                              self.DUMMY_VIF20),
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port11,
                                              self.DUMMY_VIF11),
                self._mock_send_expected_call(agent.EVPN,
                                              dummy_port21,
                                              self.DUMMY_VIF21)
            ]

            # Attach 2 ports on network 1
            self.agent.bagpipe_port_attach(None, dummy_port10)
            self.agent.bagpipe_port_attach(None, dummy_port11)

            # Attach 2 ports on network 2
            self.agent.bagpipe_port_attach(None, dummy_port20)
            self.agent.bagpipe_port_attach(None, dummy_port21)

            # Detach 1 port from each network
            self.agent.bagpipe_port_detach(None, dummy_detach10)
            self.agent.bagpipe_port_detach(None, dummy_detach20)

            # Verify attachments list consistency
            for network_id, evpn_rt in [
                    (base.NETWORK1['id'], base.BAGPIPE_L2_RT1),
                    (base.NETWORK2['id'], base.BAGPIPE_L2_RT2)]:
                self._check_network_info(network_id,
                                         1,
                                         agent.BAGPIPE_L2_SERVICE,
                                         agent.EVPN,
                                         evpn_rt)

            # Detach remaining port from each network
            self.agent.bagpipe_port_detach(None, dummy_detach11)
            self.agent.bagpipe_port_detach(None, dummy_detach21)

            # Check if calls on BaGPipe BGP API are as expected
            send_detach_fn.assert_has_calls(expected_calls)

            # Verify attachments list consistency
            for network_id in [base.NETWORK1['id'], base.NETWORK2['id']]:
                self._check_network_info(network_id, 0)

            self.assertEqual(0, len(self.agent.networks_info),
                             "Registered attachments list must be empty: %s" %
                             self.agent.networks_info)


class TestServiceBaGPipeL2LinuxBridge(
        base.BaseTestBaGPipeBGPAgentLinuxBridge, TestServiceBaGPipeL2Mixin):
    pass
